#!/usr/bin/env python

# Copyright INRA (Institut National de la Recherche Agronomique)
# http://www.inra.fr
# http://urgi.versailles.inra.fr
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import sys
from pyRepetUnit.commons.utils.RepetOptionParser import RepetOptionParser
from pyRepetUnit.commons.utils.FileUtils import FileUtils
from pyRepetUnit.commons.sql.DbFactory import DbFactory
from pyRepetUnit.commons.sql.TablePathAdaptator import TablePathAdaptator
from pyRepetUnit.commons.sql.TableSeqAdaptator import TableSeqAdaptator
from pyRepetUnit.commons.seq.Bioseq import Bioseq

class FilterMatchWithUnknownNt(object):

    def __init__(self, inPathTableName = "", inSeqTableName = "", configFileName = "TEdenovo.cfg", threshold = 90, verbose = 0):
        self._inPathTableName = inPathTableName
        self._inSeqTableName = inSeqTableName
        self._configFileName = configFileName
        self._threshold = threshold
        self._verbose = verbose

    def setAttributesFromCmdLine(self):
        parser = RepetOptionParser()
        parser.add_option("-p", "--path",    dest = "inPathTableName",   action = "store",     type = "string", help = "input path table name",                               default = "")
        parser.add_option("-s", "--seq",     dest = "inSeqTableName",    action = "store",     type = "string", help = "input seq table name",                                default = "")
        parser.add_option("-t", "--thresh",  dest = "threshPercent",     action = "store",     type = int,      help = "maximum proportion of 'N' nucleotide to filter path", default = 90)
        parser.add_option("-C", "--conf",    dest = "configFileName",    action = "store",     type = "string", help = "config file to connect to DB",                        default = "")
        parser.add_option("-v", "--verbose", dest = "verbose",           action = "store",     type = "int",    help = "verbosity level (default=0/1)",                       default = 0)
        (self._options, args) = parser.parse_args()
        self._setAttributesFromOptions(self._options)
        
    def _setAttributesFromOptions(self, options):
        self.setInPathTableName(self._options.inPathTableName)
        self.setInSeqTableName(self._options.inSeqTableName)
        self.setInConfigFileName(self._options.configFileName)
        self.setThreshold(self._options.threshPercent)
        self.setVerbose(self._options.verbose)
        
    def setInPathTableName(self, inPathTableName):
        self._inPathTableName = inPathTableName
        
    def setInSeqTableName(self, inSeqTableName):
        self._inSeqTableName = inSeqTableName
        
    def setInConfigFileName(self, configFileName):
        self._configFileName = configFileName
        
    def setThreshold(self, threshPercent):
        self._threshold = threshPercent
        
    def setVerbose(self, verbose):
        self._verbose = verbose
   
    def checkOptions(self):
        if self._configFileName != "":
            if not FileUtils.isRessourceExists(self._configFileName):
                raise Exception("ERROR: Configuration file does not exist!")
        else:
            raise Exception("ERROR: No specified -C option!")
                      
    def run(self):
        if self._verbose > 0:
            msg = "START FilterMatchWithUnknownNt"
            msg += "\ninput path table: %s" % self._inPathTableName
            msg += "\ninput seq table: %s" % self._inSeqTableName
            sys.stdout.write("%s\n" % msg)
            sys.stdout.flush()
        
        self._iDb = DbFactory.createInstance()
        iTablePathAdaptator = TablePathAdaptator(self._iDb, self._inPathTableName)
        lPathId = iTablePathAdaptator.getIdList()
        iTableSeqAdaptator = TableSeqAdaptator(self._iDb, self._inSeqTableName)
        nbRemovedMatches = 0
        
        for id in lPathId:
            lPath = iTablePathAdaptator.getPathListFromId(id)
            for path in lPath:
                qname = path.getQueryName()
                start = path.getQueryStart()
                end = path.getQueryEnd()
                lAccession = iTableSeqAdaptator.getAccessionsList()
                accession = lAccession[lAccession.index(qname)]
                seq = iTableSeqAdaptator.getSubSequence(accession, start, end)
                
                iBioseq = Bioseq("bioseq", seq)
                
                proportion = iBioseq.countNt("N")/float(iBioseq.getLength())
                
                if proportion*100 >= self._threshold:
                    nbRemovedMatches += 1
                    iTablePathAdaptator.deleteFromPath(path)
                    
        if self._verbose > 0:
            msg = "\tNumber of removed matches: %i\n" % nbRemovedMatches
            msg += "END FilterMatchWithUnknownNt\n"
            sys.stdout.write(msg)
            sys.stdout.flush()

if __name__ == '__main__':
    iFilterMatchWithUnknownNt = FilterMatchWithUnknownNt()
    iFilterMatchWithUnknownNt.setAttributesFromCmdLine()
    iFilterMatchWithUnknownNt.run() 
    