#!/usr/bin/env python

# Copyright INRA (Institut National de la Recherche Agronomique)
# http://www.inra.fr
# http://urgi.versailles.inra.fr
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

## @file
# usage: GetClassifUniq.py [ options ]
# options:
#      -h: this help
#      -p: project name
#      -v: verbosity level (default=0/1/2)

import os
import shutil
import sys
from pyRepetUnit.commons.seq.BioseqDB import BioseqDB
from pyRepetUnit.commons.utils.RepetOptionParser import RepetOptionParser
from pyRepetUnit.commons.utils.FileUtils import FileUtils

"""
Parse the 'classif' file with all the consensus before removing the redundancy and keep only the lines of the non-redundant consensus.
"""

class GetClassifUniq(object):
    
    def __init__(self, inFastaFileName, inClassifFileName, verbose = 0):
        self._options = None
        self._inClassifFileName = inClassifFileName
        self._inFastaFileName = inFastaFileName
        self._verbose = verbose
    
    def setAttributesFromCmdLine(self):
        description = "TODO"
        epilog = "TODO\n"
        
        parser = RepetOptionParser(description = description, epilog = epilog)
        parser.add_option("-i", "--classif",    dest = "inClassifFileName",   action = "store",     type = "string", help = "input classif file",                               default = "")
        parser.add_option("-j", "--fasta",      dest = "inFastaFileName",     action = "store",     type = "string", help = "input fasta file",                                 default = "")
        parser.add_option("-v", "--verbose",    dest = "verbose",             action = "store",     type = "int",    help = "verbosity level (default=0/1/2)",                  default = 0)
        (self._options, args) = parser.parse_args()
        self._setAttributesFromOptions(self._options)
        
    def _setAttributesFromOptions(self, options):
        self.setInClassifFileName(self._options.inClassifFileName)
        self.setInFastaFileName(self._options.inFastaFileName)
        self.setVerbose(self._options.verbose)
        
    def setInClassifFileName(self, inClassifFileName):
        self._inClassifFileName = inClassifFileName
        
    def setInFastaFileName(self, inFastaFileName):
        self._inFastaFileName = inFastaFileName
        
    def setVerbose(self, verbose):
        self._verbose = verbose
   
    def checkOptions(self):
        if self._inClassifFileName != "":
            if not FileUtils.isRessourceExists(self._inClassifFileName):
                raise Exception("ERROR: Input classif file does not exist!")
        else:
            raise Exception("ERROR: No specified -i option!")
        
        if self._inFastaFileName != "":
            if not FileUtils.isRessourceExists(self._inFastaFileName):
                raise Exception("ERROR: Input fasta file does not exist!")
        else:
            raise Exception("ERROR: No specified -j option!")
                      
    def run(self):
        outClassifFileName = "%s_withoutRedundancy.classif" % os.path.splitext(os.path.basename(self._inClassifFileName))[0]
            
        if self._verbose > 0:
            msg = "START %s" % (sys.argv[0].split("/")[-1])
            msg += "\ninput classification file: %s" % self._inClassifFileName
            msg += "\ninput fasta file: %s" % self._inFastaFileName
            msg += "\noutput file: %s" % outClassifFileName
            sys.stdout.write("%s\n" % msg)
            sys.stdout.flush()
            
        uniqDB = BioseqDB(self._inFastaFileName)
    
        inClassifFile = open(self._inClassifFileName, "r")
        dName2Classif = {}
        line = inClassifFile.readline()
        while True:
            if line == "":
                break
            data = line.split("\t")
            name = data[0]
            if not dName2Classif.has_key(name):
                dName2Classif[name] = line
            else:
                print "*ERROR: two consensus have the same name"
                sys.exit(1)
            line = inClassifFile.readline()
        inClassifFile.close()
    
        if os.path.exists(outClassifFileName):
            os.remove(outClassifFileName)
        outClassifFile = open(outClassifFileName, "w")
        for bs in uniqDB.db:
            outClassifFile.write(dName2Classif[bs.header])
        outClassifFile.close()
        FileUtils.sortFileContent(outClassifFileName, "tmp")
        shutil.move("tmp", outClassifFileName)
        
        if self._verbose > 0:
            msg = "END %s" % (sys.argv[0].split("/")[-1])
            sys.stdout.write("%s\n" % msg)
            sys.stdout.flush()
            
        return 0

if __name__ == '__main__':
    iGetClassifUniq = GetClassifUniq()
    iGetClassifUniq.setAttributesFromCmdLine()
    iGetClassifUniq.run() 
    